@@ -43,18 +43,12 @@
 #include <libxml/parser.h>
 #include <libxml/parserInternals.h>
 
+#include <gio/gio.h>
+
 #include <math.h>
 #include <string.h>
 #include <stdarg.h>
 
-#ifdef HAVE_GSF
-#include <gsf/gsf-input-gzip.h>
-#include <gsf/gsf-input-memory.h>
-#include <gsf/gsf-output-memory.h>
-#include <gsf/gsf-utils.h>
-#endif
-
-#include "rsvg-bpath-util.h"
 #include "rsvg-path.h"
 #include "rsvg-paint-server.h"
 #include "rsvg-xml.h"
@@ -82,10 +76,18 @@ typedef struct _RsvgSaxHandlerStyle {
     GString *style;
 } RsvgSaxHandlerStyle;
 
+typedef struct {
+    RsvgSaxHandler super;
+    RsvgHandle *ctx;
+    const char *name;
+    GString *string;
+    GString **stringptr;
+} RsvgSaxHandlerExtra;
+
 /* hide this fact from the general public */
-typedef RsvgSaxHandlerDefs RsvgSaxHandlerTitle;
-typedef RsvgSaxHandlerDefs RsvgSaxHandlerDesc;
-typedef RsvgSaxHandlerDefs RsvgSaxHandlerMetadata;
+typedef RsvgSaxHandlerExtra RsvgSaxHandlerTitle;
+typedef RsvgSaxHandlerExtra RsvgSaxHandlerDesc;
+typedef RsvgSaxHandlerExtra RsvgSaxHandlerMetadata;
 
 static void
 rsvg_style_handler_free (RsvgSaxHandler * self)
@@ -127,7 +129,7 @@ rsvg_style_handler_end (RsvgSaxHandler * self, const char *name)
 }
 
 static void
-rsvg_start_style (RsvgHandle * ctx, RsvgPropertyBag * atts)
+rsvg_start_style (RsvgHandle * ctx)
 {
     RsvgSaxHandlerStyle *handler = g_new0 (RsvgSaxHandlerStyle, 1);
 
@@ -280,22 +282,35 @@ rsvg_standard_element_start (RsvgHandle * ctx, const char *name, RsvgPropertyBag
     }
 }
 
-/* start desc */
+/* extra (title, desc, metadata) */
 
 static void
-rsvg_desc_handler_free (RsvgSaxHandler * self)
+rsvg_extra_handler_free (RsvgSaxHandler * self)
 {
+    RsvgSaxHandlerExtra *z = (RsvgSaxHandlerExtra *) self;
+
+    if (z->stringptr) {
+        if (*z->stringptr)
+            g_string_free (*z->stringptr, TRUE);
+        *z->stringptr = z->string;
+    } else if (z->string) {
+        g_string_free (z->string, TRUE);
+    }
+
     g_free (self);
 }
 
 static void
-rsvg_desc_handler_characters (RsvgSaxHandler * self, const char *ch, int len)
+rsvg_extra_handler_characters (RsvgSaxHandler * self, const char *ch, int len)
 {
-    RsvgSaxHandlerDesc *z = (RsvgSaxHandlerDesc *) self;
+    RsvgSaxHandlerExtra *z = (RsvgSaxHandlerExtra *) self;
     RsvgHandle *ctx = z->ctx;
 
     /* This isn't quite the correct behavior - in theory, any graphics
-       element may contain a title or desc element */
+       element may contain a title, desc, or metadata element */
+
+    if (!z->string)
+        return;
 
     if (!ch || !len)
         return;
@@ -303,25 +318,25 @@ rsvg_desc_handler_characters (RsvgSaxHandler * self, const char *ch, int len)
     if (!g_utf8_validate ((char *) ch, len, NULL)) {
         char *utf8;
         utf8 = rsvg_make_valid_utf8 ((char *) ch, len);
-        g_string_append (ctx->priv->desc, utf8);
+        g_string_append (z->string, utf8);
         g_free (utf8);
     } else {
-        g_string_append_len (ctx->priv->desc, (char *) ch, len);
+        g_string_append_len (z->string, (char *) ch, len);
     }
 }
 
 static void
-rsvg_desc_handler_start (RsvgSaxHandler * self, const char *name, RsvgPropertyBag * atts)
+rsvg_extra_handler_start (RsvgSaxHandler * self, const char *name, RsvgPropertyBag * atts)
 {
 }
 
 static void
-rsvg_desc_handler_end (RsvgSaxHandler * self, const char *name)
+rsvg_extra_handler_end (RsvgSaxHandler * self, const char *name)
 {
-    RsvgSaxHandlerDesc *z = (RsvgSaxHandlerDesc *) self;
+    RsvgSaxHandlerExtra *z = (RsvgSaxHandlerExtra *) self;
     RsvgHandle *ctx = z->ctx;
 
-    if (!strcmp (name, "desc")) {
+    if (!strcmp (name, z->name)) {
         if (ctx->priv->handler != NULL) {
             ctx->priv->handler->free (ctx->priv->handler);
             ctx->priv->handler = NULL;
@@ -329,19 +344,42 @@ rsvg_desc_handler_end (RsvgSaxHandler * self, const char *name)
     }
 }
 
-static void
-rsvg_start_desc (RsvgHandle * ctx, RsvgPropertyBag * atts)
+static RsvgSaxHandlerExtra *
+rsvg_start_extra (RsvgHandle * ctx,
+                  const char *name,
+                  GString **stringptr)
 {
-    RsvgSaxHandlerDesc *handler = g_new0 (RsvgSaxHandlerDesc, 1);
+    RsvgSaxHandlerExtra *handler = g_new0 (RsvgSaxHandlerExtra, 1);
+    RsvgNode *treebase = ctx->priv->treebase;
+    RsvgNode *currentnode = ctx->priv->currentnode;
+    gboolean do_care;
 
-    handler->super.free = rsvg_desc_handler_free;
-    handler->super.characters = rsvg_desc_handler_characters;
-    handler->super.start_element = rsvg_desc_handler_start;
-    handler->super.end_element = rsvg_desc_handler_end;
+    /* only parse <extra> for the <svg> node.
+     * This isn't quite the correct behavior - any graphics
+     * element may contain a <extra> element.
+     */
+    do_care = treebase != NULL && treebase == currentnode;
+
+    handler->super.free = rsvg_extra_handler_free;
+    handler->super.characters = do_care ? rsvg_extra_handler_characters : NULL;
+    handler->super.start_element = rsvg_extra_handler_start;
+    handler->super.end_element = rsvg_extra_handler_end;
     handler->ctx = ctx;
+    handler->name = name; /* interned */
+    handler->string = do_care ? g_string_new (NULL) : NULL;
+    handler->stringptr = do_care ? stringptr : NULL;
 
-    ctx->priv->desc = g_string_new (NULL);
     ctx->priv->handler = &handler->super;
+
+    return handler;
+}
+
+/* start desc */
+
+static void
+rsvg_start_desc (RsvgHandle * ctx)
+{
+    rsvg_start_extra (ctx, "desc", &ctx->priv->desc);
 }
 
 /* end desc */
@@ -349,99 +387,15 @@ rsvg_start_desc (RsvgHandle * ctx, RsvgPropertyBag * atts)
 /* start title */
 
 static void
-rsvg_title_handler_free (RsvgSaxHandler * self)
+rsvg_start_title (RsvgHandle * ctx)
 {
-    g_free (self);
-}
-
-static void
-rsvg_title_handler_characters (RsvgSaxHandler * self, const char *ch, int len)
-{
-    RsvgSaxHandlerTitle *z = (RsvgSaxHandlerTitle *) self;
-    RsvgHandle *ctx = z->ctx;
-
-    /* This isn't quite the correct behavior - in theory, any graphics
-       element may contain a title or desc element */
-
-    if (!ch || !len)
-        return;
-
-    if (!g_utf8_validate ((char *) ch, len, NULL)) {
-        char *utf8;
-        utf8 = rsvg_make_valid_utf8 ((char *) ch, len);
-        g_string_append (ctx->priv->title, utf8);
-        g_free (utf8);
-    } else {
-        g_string_append_len (ctx->priv->title, (char *) ch, len);
-    }
-}
-
-static void
-rsvg_title_handler_start (RsvgSaxHandler * self, const char *name, RsvgPropertyBag * atts)
-{
-}
-
-static void
-rsvg_title_handler_end (RsvgSaxHandler * self, const char *name)
-{
-    RsvgSaxHandlerTitle *z = (RsvgSaxHandlerTitle *) self;
-    RsvgHandle *ctx = z->ctx;
-
-    if (!strcmp (name, "title")) {
-        if (ctx->priv->handler != NULL) {
-            ctx->priv->handler->free (ctx->priv->handler);
-            ctx->priv->handler = NULL;
-        }
-    }
-}
-
-static void
-rsvg_start_title (RsvgHandle * ctx, RsvgPropertyBag * atts)
-{
-    RsvgSaxHandlerTitle *handler = g_new0 (RsvgSaxHandlerTitle, 1);
-
-    handler->super.free = rsvg_title_handler_free;
-    handler->super.characters = rsvg_title_handler_characters;
-    handler->super.start_element = rsvg_title_handler_start;
-    handler->super.end_element = rsvg_title_handler_end;
-    handler->ctx = ctx;
-
-    ctx->priv->title = g_string_new (NULL);
-    ctx->priv->handler = &handler->super;
+    rsvg_start_extra (ctx, "title", &ctx->priv->title);
 }
 
 /* end title */
 
 /* start metadata */
 
-static void
-rsvg_metadata_handler_free (RsvgSaxHandler * self)
-{
-    g_free (self);
-}
-
-static void
-rsvg_metadata_handler_characters (RsvgSaxHandler * self, const char *ch, int len)
-{
-    RsvgSaxHandlerDesc *z = (RsvgSaxHandlerDesc *) self;
-    RsvgHandle *ctx = z->ctx;
-
-    /* This isn't quite the correct behavior - in theory, any graphics
-       element may contain a metadata or desc element */
-
-    if (!ch || !len)
-        return;
-
-    if (!g_utf8_validate ((char *) ch, len, NULL)) {
-        char *utf8;
-        utf8 = rsvg_make_valid_utf8 ((char *) ch, len);
-        g_string_append (ctx->priv->metadata, utf8);
-        g_free (utf8);
-    } else {
-        g_string_append_len (ctx->priv->metadata, (char *) ch, len);
-    }
-}
-
 static void
 rsvg_metadata_props_enumerate (const char *key, const char *value, gpointer user_data)
 {
@@ -455,9 +409,14 @@ rsvg_metadata_handler_start (RsvgSaxHandler * self, const char *name, RsvgProper
     RsvgSaxHandlerMetadata *z = (RsvgSaxHandlerMetadata *) self;
     RsvgHandle *ctx = z->ctx;
 
-    g_string_append_printf (ctx->priv->metadata, "<%s ", name);
-    rsvg_property_bag_enumerate (atts, rsvg_metadata_props_enumerate, ctx->priv->metadata);
-    g_string_append (ctx->priv->metadata, ">\n");
+    rsvg_extra_handler_start (self, name, atts);
+
+    if (!z->string)
+        return;
+
+    g_string_append_printf (z->string, "<%s ", name);
+    rsvg_property_bag_enumerate (atts, rsvg_metadata_props_enumerate, z->string);
+    g_string_append (z->string, ">\n");
 }
 
 static void
@@ -466,28 +425,21 @@ rsvg_metadata_handler_end (RsvgSaxHandler * self, const char *name)
     RsvgSaxHandlerMetadata *z = (RsvgSaxHandlerMetadata *) self;
     RsvgHandle *ctx = z->ctx;
 
-    if (!strcmp (name, "metadata")) {
-        if (ctx->priv->handler != NULL) {
-            ctx->priv->handler->free (ctx->priv->handler);
-            ctx->priv->handler = NULL;
-        }
-    } else
-        g_string_append_printf (ctx->priv->metadata, "</%s>\n", name);
+    if (strcmp (name, z->name) != 0) {
+        if (z->string)
+            g_string_append_printf (z->string, "</%s>\n", name);
+    } else {
+        rsvg_extra_handler_end (self, name);
+    }
 }
 
 static void
-rsvg_start_metadata (RsvgHandle * ctx, RsvgPropertyBag * atts)
+rsvg_start_metadata (RsvgHandle * ctx)
 {
-    RsvgSaxHandlerMetadata *handler = g_new0 (RsvgSaxHandlerMetadata, 1);
+    RsvgSaxHandlerMetadata *handler = rsvg_start_extra (ctx, "metadata", &ctx->priv->metadata);
 
-    handler->super.free = rsvg_metadata_handler_free;
-    handler->super.characters = rsvg_metadata_handler_characters;
     handler->super.start_element = rsvg_metadata_handler_start;
     handler->super.end_element = rsvg_metadata_handler_end;
-    handler->ctx = ctx;
-
-    ctx->priv->metadata = g_string_new (NULL);
-    ctx->priv->handler = &handler->super;
 }
 
 /* end metadata */
@@ -602,11 +554,10 @@ rsvg_start_xinclude (RsvgHandle * ctx, RsvgPropertyBag * atts)
                 /* xml */
                 xmlDocPtr xml_doc;
                 xmlParserCtxtPtr xml_parser;
-                int result;
 
                 xml_parser = xmlCreatePushParserCtxt (&rsvgSAXHandlerStruct, ctx, NULL, 0, NULL);
-                result = xmlParseChunk (xml_parser, (char *) data->data, data->len, 0);
-                result = xmlParseChunk (xml_parser, "", 0, TRUE);
+                (void) xmlParseChunk (xml_parser, (char *) data->data, data->len, 0);
+                (void) xmlParseChunk (xml_parser, "", 0, TRUE);
 
                 xml_doc = xml_parser->myDoc;
                 xmlFreeParserCtxt (xml_parser);
@@ -653,13 +604,13 @@ rsvg_start_element (void *data, const xmlChar * name, const xmlChar ** atts)
                 name = (const xmlChar *) (tempname + 1);
 
         if (!strcmp ((const char *) name, "style"))
-            rsvg_start_style (ctx, bag);
+            rsvg_start_style (ctx);
         else if (!strcmp ((const char *) name, "title"))
-            rsvg_start_title (ctx, bag);
+            rsvg_start_title (ctx);
         else if (!strcmp ((const char *) name, "desc"))
-            rsvg_start_desc (ctx, bag);
+            rsvg_start_desc (ctx);
         else if (!strcmp ((const char *) name, "metadata"))
-            rsvg_start_metadata (ctx, bag);
+            rsvg_start_metadata (ctx);
         else if (!strcmp ((const char *) name, "include"))      /* xi:include */
             rsvg_start_xinclude (ctx, bag);
         else
@@ -753,6 +704,9 @@ rsvg_characters_impl (RsvgHandle * ctx, const xmlChar * ch, int len)
                 if (RSVG_NODE_TYPE (node) == RSVG_NODE_TYPE_CHARS) {
                     self = (RsvgNodeChars*)node;
                 }
+                else if (RSVG_NODE_TYPE (node) == RSVG_NODE_TYPE_TSPAN) {
+                    self = NULL;
+                }
             }
 
             if (self != NULL) {
@@ -790,14 +744,6 @@ rsvg_characters (void *data, const xmlChar * ch, int len)
     rsvg_characters_impl (ctx, ch, len);
 }
 
-#if LIBXML_VERSION >= 20621
-#define RSVG_ENABLE_ENTITIES
-#elif defined(__GNUC__)
-#warning "libxml version less than 2.6.22. XML entities won't work"
-#endif
-
-#ifdef RSVG_ENABLE_ENTITIES
-
 static xmlEntityPtr
 rsvg_get_entity (void *data, const xmlChar * name)
 {
@@ -816,7 +762,6 @@ rsvg_entity_decl (void *data, const xmlChar * name, int type,
     RsvgHandle *ctx = (RsvgHandle *) data;
     GHashTable *entities = ctx->priv->entities;
     xmlEntityPtr entity;
-#if LIBXML_VERSION >= 20700
     xmlChar *resolvedSystemId = NULL, *resolvedPublicId = NULL;
 
     if (systemId)
@@ -845,46 +790,6 @@ rsvg_entity_decl (void *data, const xmlChar * name, int type,
     xmlFree(resolvedSystemId);
 
     g_hash_table_insert (entities, g_strdup ((const char*) name), entity);
-#else
-    xmlChar *dupname;
-
-    entity = xmlMalloc (sizeof (xmlEntity));
-    memset (entity, 0, sizeof (xmlEntity));
-    entity->type = XML_ENTITY_DECL;
-    dupname = (xmlChar *) xmlMemStrdup ((const char *) name);
-    entity->name = dupname;
-
-    entity->etype = type;
-    if (content) {
-        entity->content = (xmlChar *) xmlMemStrdup ((const char *) content);
-        entity->length = strlen ((const char *) content);
-    } else if (systemId || publicId) {
-        GByteArray *data = NULL;
-
-        if (systemId)
-            data =
-                _rsvg_acquire_xlink_href_resource ((const char *) systemId,
-                                                   rsvg_handle_get_base_uri (ctx), NULL);
-        else if (publicId)
-            data =
-                _rsvg_acquire_xlink_href_resource ((const char *) publicId,
-                                                   rsvg_handle_get_base_uri (ctx), NULL);
-
-        if (data) {
-            entity->SystemID = (xmlChar *) xmlMemStrdup ((const char *) systemId);
-            entity->ExternalID = (xmlChar *) xmlMemStrdup ((const char *) publicId);
-            entity->content = (xmlChar *) xmlMemStrdup ((const char *) data->data);
-            entity->length = data->len;
-
-            /* fool libxml2 into supporting SYSTEM and PUBLIC entities */
-            entity->etype = XML_INTERNAL_GENERAL_ENTITY;
-
-            g_byte_array_free (data, TRUE);
-        }
-    }
-
-    g_hash_table_insert (entities, dupname, entity);
-#endif
 }
 
 static void
@@ -907,8 +812,6 @@ rsvg_get_parameter_entity (void *data, const xmlChar * name)
     return entity;
 }
 
-#endif
-
 static void
 rsvg_error_cb (void *data, const char *msg, ...)
 {
@@ -932,17 +835,17 @@ rsvg_processing_instruction (void *ctx, const xmlChar * target, const xmlChar *
         char **xml_atts;
 
         xml_atts = rsvg_css_parse_xml_attribute_string ((const char *) data);
-        atts = rsvg_property_bag_new ((const char **) xml_atts);
 
-        if (atts) {
+        if (xml_atts) {
             const char *value;
 
+            atts = rsvg_property_bag_new ((const char **) xml_atts);
             value = rsvg_property_bag_lookup (atts, "alternate");
-            if (!value || (strcmp (value, "no") != 0)) {
+            if (!value || !value[0] || (strcmp (value, "no") != 0)) {
                 value = rsvg_property_bag_lookup (atts, "type");
                 if (value && strcmp (value, "text/css") == 0) {
                     value = rsvg_property_bag_lookup (atts, "href");
-                    if (value) {
+                    if (value && value[0]) {
                         GByteArray *style;
 
                         style =
@@ -957,8 +860,8 @@ rsvg_processing_instruction (void *ctx, const xmlChar * target, const xmlChar *
                 }
             }
 
-            g_strfreev (xml_atts);
             rsvg_property_bag_free (atts);
+            g_strfreev (xml_atts);
         }
     }
 }
@@ -971,12 +874,10 @@ rsvg_SAX_handler_struct_init (void)
 
         memset (&rsvgSAXHandlerStruct, 0, sizeof (rsvgSAXHandlerStruct));
 
-#ifdef RSVG_ENABLE_ENTITIES
         rsvgSAXHandlerStruct.getEntity = rsvg_get_entity;
         rsvgSAXHandlerStruct.entityDecl = rsvg_entity_decl;
         rsvgSAXHandlerStruct.unparsedEntityDecl = rsvg_unparsed_entity_decl;
         rsvgSAXHandlerStruct.getParameterEntity = rsvg_get_parameter_entity;
-#endif
         rsvgSAXHandlerStruct.characters = rsvg_characters;
         rsvgSAXHandlerStruct.error = rsvg_error_cb;
         rsvgSAXHandlerStruct.cdataBlock = rsvg_characters;
@@ -1259,6 +1160,8 @@ rsvg_drawing_ctx_free (RsvgDrawingCtx * handle)
  * Returns: The SVG's title
  *
  * Since: 2.9
+ *
+ * Deprecated: 2.36
  */
 const char *
 rsvg_handle_get_metadata (RsvgHandle * handle)
@@ -1281,6 +1184,8 @@ rsvg_handle_get_metadata (RsvgHandle * handle)
  * Returns: The SVG's title
  *
  * Since: 2.4
+ *
+ * Deprecated: 2.36
  */
 const char *
 rsvg_handle_get_title (RsvgHandle * handle)
@@ -1303,6 +1208,8 @@ rsvg_handle_get_title (RsvgHandle * handle)
  * Returns: The SVG's description
  *
  * Since: 2.4
+ *
+ * Deprecated: 2.36
  */
 const char *
 rsvg_handle_get_desc (RsvgHandle * handle)
@@ -1359,7 +1266,6 @@ rsvg_handle_get_dimensions_sub (RsvgHandle * handle, RsvgDimensionData * dimensi
     cairo_t *cr;
     cairo_surface_t *target;
     RsvgDrawingCtx *draw;
-    RsvgCairoRender *render;
     RsvgNodeSvg *root = NULL;
     RsvgNode *sself = NULL;
     RsvgBbox bbox;
@@ -1388,8 +1294,8 @@ rsvg_handle_get_dimensions_sub (RsvgHandle * handle, RsvgDimensionData * dimensi
     if (!root)
         return FALSE;
 
-    bbox.x = bbox.y = 0;
-    bbox.w = bbox.h = 1;
+    bbox.rect.x = bbox.rect.y = 0;
+    bbox.rect.width = bbox.rect.height = 1;
 
     if (!id && (root->w.factor == 'p' || root->h.factor == 'p')
             && !root->vbox.active)
@@ -1420,12 +1326,7 @@ rsvg_handle_get_dimensions_sub (RsvgHandle * handle, RsvgDimensionData * dimensi
         cairo_save (cr);
 
         rsvg_node_draw ((RsvgNode *) handle->priv->treebase, draw, 0);
-        render = (RsvgCairoRender *) draw->render;
-
-        bbox.x = render->bbox.x;
-        bbox.y = render->bbox.y;
-        bbox.w = render->bbox.w;
-        bbox.h = render->bbox.h;
+        bbox = RSVG_CAIRO_RENDER (draw->render)->bbox;
 
         cairo_restore (cr);
         rsvg_state_pop (draw);
@@ -1433,16 +1334,16 @@ rsvg_handle_get_dimensions_sub (RsvgHandle * handle, RsvgDimensionData * dimensi
         cairo_destroy (cr);
         cairo_surface_destroy (target);
 
-        dimension_data->width = bbox.w;
-        dimension_data->height = bbox.h;
+        dimension_data->width = bbox.rect.width;
+        dimension_data->height = bbox.rect.height;
     } else {
-        bbox.w = root->vbox.w;
-        bbox.h = root->vbox.h;
+        bbox.rect.width = root->vbox.rect.width;
+        bbox.rect.height = root->vbox.rect.height;
 
         dimension_data->width = (int) (_rsvg_css_hand_normalize_length (&root->w, handle->priv->dpi_x,
-                                       bbox.w + bbox.x * 2, 12) + 0.5);
+                                       bbox.rect.width + bbox.rect.x * 2, 12) + 0.5);
         dimension_data->height = (int) (_rsvg_css_hand_normalize_length (&root->h, handle->priv->dpi_y,
-                                         bbox.h + bbox.y * 2,
+                                         bbox.rect.height + bbox.rect.y * 2,
                                          12) + 0.5);
     }
     
@@ -1523,19 +1424,16 @@ rsvg_handle_get_position_sub (RsvgHandle * handle, RsvgPositionData * position_d
     cairo_save (cr);
 
     rsvg_node_draw ((RsvgNode *) handle->priv->treebase, draw, 0);
-    bbox.x = ((RsvgCairoRender *) draw->render)->bbox.x;
-    bbox.y = ((RsvgCairoRender *) draw->render)->bbox.y;
-    bbox.w = ((RsvgCairoRender *) draw->render)->bbox.w;
-    bbox.h = ((RsvgCairoRender *) draw->render)->bbox.h;
+    bbox = RSVG_CAIRO_RENDER (draw->render)->bbox;
 
     cairo_restore (cr);
     rsvg_state_pop (draw);
     rsvg_drawing_ctx_free (draw);
 
-    position_data->x = bbox.x;
-    position_data->y = bbox.y;
-    dimension_data.width = bbox.w;
-    dimension_data.height = bbox.h;
+    position_data->x = bbox.rect.x;
+    position_data->y = bbox.rect.y;
+    dimension_data.width = bbox.rect.width;
+    dimension_data.height = bbox.rect.height;
 
     dimension_data.em = dimension_data.width;
     dimension_data.ex = dimension_data.height;
@@ -1725,28 +1623,15 @@ rsvg_handle_write (RsvgHandle * handle, const guchar * buf, gsize count, GError
         /* test for GZ marker. todo: store the first 2 bytes in the odd circumstance that someone calls
          * write() in 1 byte increments */
         if ((count >= 2) && (buf[0] == (guchar) 0x1f) && (buf[1] == (guchar) 0x8b)) {
-#if GLIB_CHECK_VERSION (2, 24, 0)
             priv->data_input_stream = g_memory_input_stream_new ();
-#elif defined (HAVE_GSF)
-            priv->gzipped_data = GSF_OUTPUT (gsf_output_memory_new ());
-#else
-            g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED,
-                                 "GZip compressed SVG not supported");
-            return FALSE;
-#endif
         }
     }
 
-#if GLIB_CHECK_VERSION (2, 24, 0)
     if (priv->data_input_stream) {
         g_memory_input_stream_add_data ((GMemoryInputStream *) priv->data_input_stream,
                                         g_memdup (buf, count), count, (GDestroyNotify) g_free);
         return TRUE;
     }
-#elif defined (HAVE_GSF)
-    if (priv->gzipped_data)
-        return gsf_output_write (handle->priv->gzipped_data, count, buf);
-#endif
 
     return rsvg_handle_write_impl (handle, buf, count, error);
 }
@@ -1773,7 +1658,6 @@ rsvg_handle_close (RsvgHandle * handle, GError ** error)
     if (priv->is_closed)
           return TRUE;
 
-#if GLIB_CHECK_VERSION (2, 24, 0)
     if (priv->data_input_stream) {
         GConverter *converter;
         GInputStream *stream;
@@ -1790,49 +1674,6 @@ rsvg_handle_close (RsvgHandle * handle, GError ** error)
 
         return ret;
     }
-#elif defined(HAVE_GSF)
-    if (priv->gzipped_data) {
-        GsfInput *gzip;
-        const guchar *bytes;
-        gsize size;
-        gsize remaining;
-
-        bytes = gsf_output_memory_get_bytes (GSF_OUTPUT_MEMORY (priv->gzipped_data));
-        size = gsf_output_size (priv->gzipped_data);
-
-        gzip =
-            GSF_INPUT (gsf_input_gzip_new
-                       (GSF_INPUT (gsf_input_memory_new (bytes, size, FALSE)), error));
-        remaining = gsf_input_remaining (gzip);
-        while ((size = MIN (remaining, 1024)) > 0) {
-            guint8 const *buf;
-
-            /* write to parent */
-            buf = gsf_input_read (gzip, size, NULL);
-            if (!buf) {
-                /* an error occured, so bail */
-                g_warning (_("rsvg_gz_handle_close_impl: gsf_input_read returned NULL"));
-                break;
-            }
-
-            rsvg_handle_write_impl (handle, buf, size, error);
-            /* if we didn't manage to lower remaining number of bytes,
-             * something is wrong, and we should avoid an endless loop */
-            if (remaining == ((gsize) gsf_input_remaining (gzip))) {
-                g_warning (_
-                           ("rsvg_gz_handle_close_impl: write_impl didn't lower the input_remaining count"));
-                break;
-            }
-            remaining = gsf_input_remaining (gzip);
-        }
-        g_object_unref (gzip);
-
-        /* close parent */
-        gsf_output_close (priv->gzipped_data);
-        g_object_unref (priv->gzipped_data);
-        priv->gzipped_data = NULL;
-    }
-#endif /* GIO >= 2.24.0 */
 
     return rsvg_handle_close_impl (handle, error);
 }
@@ -1890,15 +1731,12 @@ rsvg_handle_read_stream_sync (RsvgHandle   *handle,
 
     buffer = _rsvg_xml_input_buffer_new_from_stream (stream, cancellable, XML_CHAR_ENCODING_NONE, &err);
     input = xmlNewIOInputStream (priv->ctxt, buffer, XML_CHAR_ENCODING_NONE);
-#if LIBXML_VERSION >= 20700
+
     if (xmlPushInput (priv->ctxt, input) < 0) {
         rsvg_set_error (error, priv->ctxt);
         xmlFreeInputStream (input);
         return FALSE;
     }
-#else
-    xmlPushInput (priv->ctxt, input);
-#endif
 
     result = xmlParseDocument (priv->ctxt);
     if (result != 0) {
@@ -2004,7 +1842,7 @@ rsvg_handle_new_from_stream_sync (GInputStream   *stream,
     g_return_val_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable), NULL);
     g_return_val_if_fail (error == NULL || *error == NULL, NULL);
 
-    handle = rsvg_handle_new ();
+    handle = rsvg_handle_new_with_flags (flags);
 
     if (base_file)
         rsvg_handle_set_base_gfile (handle, base_file);
@@ -2022,32 +1860,38 @@ rsvg_handle_new_from_stream_sync (GInputStream   *stream,
  *
  * Initializes librsvg
  * Since: 2.9
+ * Deprecated: 2.36: Use g_type_init()
  **/
 void
 rsvg_init (void)
 {
     g_type_init ();
-
-#ifdef HAVE_GSF
-    gsf_init ();
-#endif
-
-    xmlInitParser ();
 }
 
 /**
  * rsvg_term:
  *
- * De-initializes librsvg
+ * This function does nothing.
+ *
  * Since: 2.9
+ * Deprecated: 2.36
  **/
 void
 rsvg_term (void)
 {
-#ifdef HAVE_GSF
-    gsf_shutdown ();
-#endif
+}
 
+/**
+ * rsvg_cleanup:
+ *
+ * This function should not be called from normal programs.
+ * See xmlCleanupParser() for more information.
+ *
+ * Since: 2.36
+ **/
+void
+rsvg_cleanup (void)
+{
     xmlCleanupParser ();
 }
 
@@ -2070,24 +1914,19 @@ rsvg_push_discrete_layer (RsvgDrawingCtx * ctx)
 }
 
 void
-rsvg_render_path (RsvgDrawingCtx * ctx, const char *d)
+rsvg_render_path (RsvgDrawingCtx * ctx, const cairo_path_t *path)
 {
-    /* todo: store and use the bpath higher up */
-    RsvgBpathDef *bpath_def;
-
-    bpath_def = rsvg_parse_path (d);
-    rsvg_bpath_def_art_finish (bpath_def);
-
-    ctx->render->render_path (ctx, bpath_def);
-    rsvg_render_markers (bpath_def, ctx);
-
-    rsvg_bpath_def_free (bpath_def);
+    ctx->render->render_path (ctx, path);
+    rsvg_render_markers (ctx, path);
 }
 
 void
-rsvg_render_image (RsvgDrawingCtx * ctx, GdkPixbuf * pb, double x, double y, double w, double h)
+rsvg_render_surface (RsvgDrawingCtx * ctx, cairo_surface_t *surface, double x, double y, double w, double h)
 {
-    ctx->render->render_image (ctx, pb, x, y, w, h);
+    /* surface must be a cairo image surface */
+    g_return_if_fail (cairo_surface_get_type (surface) == CAIRO_SURFACE_TYPE_IMAGE);
+
+    ctx->render->render_surface (ctx, surface, x, y, w, h);
 }
 
 void
@@ -2096,10 +1935,10 @@ rsvg_add_clipping_rect (RsvgDrawingCtx * ctx, double x, double y, double w, doub
     ctx->render->add_clipping_rect (ctx, x, y, w, h);
 }
 
-GdkPixbuf *
-rsvg_get_image_of_node (RsvgDrawingCtx * ctx, RsvgNode * drawable, double w, double h)
+cairo_surface_t *
+rsvg_get_surface_of_node (RsvgDrawingCtx * ctx, RsvgNode * drawable, double w, double h)
 {
-    return ctx->render->get_image_of_node (ctx, drawable, w, h);
+    return ctx->render->get_surface_of_node (ctx, drawable, w, h);
 }
 
 void
@@ -2109,18 +1948,16 @@ rsvg_render_free (RsvgRender * render)
 }
 
 void
-rsvg_bbox_init (RsvgBbox * self, double *affine)
+rsvg_bbox_init (RsvgBbox * self, cairo_matrix_t *affine)
 {
-    int i;
     self->virgin = 1;
-    for (i = 0; i < 6; i++)
-        self->affine[i] = affine[i];
+    self->affine = *affine;
 }
 
 void
 rsvg_bbox_insert (RsvgBbox * dst, RsvgBbox * src)
 {
-    double affine[6];
+    cairo_matrix_t affine;
     double xmin, ymin;
     double xmax, ymax;
     int i;
@@ -2129,21 +1966,24 @@ rsvg_bbox_insert (RsvgBbox * dst, RsvgBbox * src)
         return;
 
     if (!dst->virgin) {
-        xmin = dst->x, ymin = dst->y;
-        xmax = dst->x + dst->w, ymax = dst->y + dst->h;
+        xmin = dst->rect.x, ymin = dst->rect.y;
+        xmax = dst->rect.x + dst->rect.width, ymax = dst->rect.y + dst->rect.height;
     } else {
         xmin = ymin = xmax = ymax = 0;
     }
 
-    _rsvg_affine_invert (affine, dst->affine);
-    _rsvg_affine_multiply (affine, src->affine, affine);
+    affine = dst->affine;
+    if (cairo_matrix_invert (&affine) != CAIRO_STATUS_SUCCESS)
+      return; //FIXMEchpe correct??
+
+    cairo_matrix_multiply (&affine, &src->affine, &affine);
 
     for (i = 0; i < 4; i++) {
         double rx, ry, x, y;
-        rx = src->x + src->w * (double) (i % 2);
-        ry = src->y + src->h * (double) (i / 2);
-        x = affine[0] * rx + affine[2] * ry + affine[4];
-        y = affine[1] * rx + affine[3] * ry + affine[5];
+        rx = src->rect.x + src->rect.width * (double) (i % 2);
+        ry = src->rect.y + src->rect.height * (double) (i / 2);
+        x = affine.xx * rx + affine.xy * ry + affine.x0;
+        y = affine.yx * rx + affine.yy * ry + affine.y0;
         if (dst->virgin) {
             xmin = xmax = x;
             ymin = ymax = y;
@@ -2159,16 +1999,16 @@ rsvg_bbox_insert (RsvgBbox * dst, RsvgBbox * src)
                 ymax = y;
         }
     }
-    dst->x = xmin;
-    dst->y = ymin;
-    dst->w = xmax - xmin;
-    dst->h = ymax - ymin;
+    dst->rect.x = xmin;
+    dst->rect.y = ymin;
+    dst->rect.width = xmax - xmin;
+    dst->rect.height = ymax - ymin;
 }
 
 void
 rsvg_bbox_clip (RsvgBbox * dst, RsvgBbox * src)
 {
-    double affine[6];
+    cairo_matrix_t affine;
 	double xmin, ymin;
 	double xmax, ymax;
     int i;
@@ -2177,21 +2017,24 @@ rsvg_bbox_clip (RsvgBbox * dst, RsvgBbox * src)
         return;
 
 	if (!dst->virgin) {
-        xmin = dst->x + dst->w, ymin = dst->y + dst->h;
-        xmax = dst->x, ymax = dst->y;
+        xmin = dst->rect.x + dst->rect.width, ymin = dst->rect.y + dst->rect.height;
+        xmax = dst->rect.x, ymax = dst->rect.y;
     } else {
         xmin = ymin = xmax = ymax = 0;
     }
 
-    _rsvg_affine_invert (affine, dst->affine);
-    _rsvg_affine_multiply (affine, src->affine, affine);
+    affine = dst->affine;
+    if (cairo_matrix_invert (&affine) != CAIRO_STATUS_SUCCESS)
+      return;
+
+    cairo_matrix_multiply (&affine, &src->affine, &affine);
 
     for (i = 0; i < 4; i++) {
         double rx, ry, x, y;
-        rx = src->x + src->w * (double) (i % 2);
-        ry = src->y + src->h * (double) (i / 2);
-        x = affine[0] * rx + affine[2] * ry + affine[4];
-        y = affine[1] * rx + affine[3] * ry + affine[5];
+        rx = src->rect.x + src->rect.width * (double) (i % 2);
+        ry = src->rect.y + src->rect.height * (double) (i / 2);
+        x = affine.xx * rx + affine.xy * ry + affine.x0;
+        y = affine.yx * rx + affine.yy * ry + affine.y0;
         if (dst->virgin) {
             xmin = xmax = x;
             ymin = ymax = y;
@@ -2208,19 +2051,19 @@ rsvg_bbox_clip (RsvgBbox * dst, RsvgBbox * src)
         }
     }
 
-    if (xmin < dst->x)
-        xmin = dst->x;
-    if (ymin < dst->y)
-        ymin = dst->y;
-    if (xmax > dst->x + dst->w)
-        xmax = dst->x + dst->w;
-    if (ymax > dst->y + dst->h)
-        ymax = dst->y + dst->h;
+    if (xmin < dst->rect.x)
+        xmin = dst->rect.x;
+    if (ymin < dst->rect.y)
+        ymin = dst->rect.y;
+    if (xmax > dst->rect.x + dst->rect.width)
+        xmax = dst->rect.x + dst->rect.width;
+    if (ymax > dst->rect.y + dst->rect.height)
+        ymax = dst->rect.y + dst->rect.height;
 
-    dst->x = xmin;
-    dst->w = xmax - xmin;
-    dst->y = ymin;
-    dst->h = ymax - ymin;
+    dst->rect.x = xmin;
+    dst->rect.width = xmax - xmin;
+    dst->rect.y = ymin;
+    dst->rect.height = ymax - ymin;
 }
 
 void
@@ -2229,8 +2072,8 @@ _rsvg_push_view_box (RsvgDrawingCtx * ctx, double w, double h)
     RsvgViewBox *vb = g_new (RsvgViewBox, 1);
     *vb = ctx->vb;
     ctx->vb_stack = g_slist_prepend (ctx->vb_stack, vb);
-    ctx->vb.w = w;
-    ctx->vb.h = h;
+    ctx->vb.rect.width = w;
+    ctx->vb.rect.height = h;
 }
 
 void
