@@ -179,54 +179,54 @@ _rsvg_acquire_xlink_href_resource (const char *href, const char *base_uri, GErro
     return arr;
 }
 
-GdkPixbuf *
-rsvg_pixbuf_new_from_href (const char *href, const char *base_uri, GError ** error)
+cairo_surface_t *
+rsvg_cairo_surface_new_from_href (const char *href, 
+                                  const char *base_uri, 
+                                  GError **error)
 {
     GByteArray *arr;
+    GdkPixbufLoader *loader;
+    GdkPixbuf *pixbuf = NULL;
+    int res;
+    cairo_surface_t *surface;
 
     arr = _rsvg_acquire_xlink_href_resource (href, base_uri, error);
-    if (arr) {
-        GdkPixbufLoader *loader;
-        GdkPixbuf *pixbuf = NULL;
-        int res;
+    if (arr == NULL)
+        return NULL;
 
-        loader = gdk_pixbuf_loader_new ();
+    loader = gdk_pixbuf_loader_new ();
 
-        res = gdk_pixbuf_loader_write (loader, arr->data, arr->len, error);
-        g_byte_array_free (arr, TRUE);
-
-        if (!res) {
-            gdk_pixbuf_loader_close (loader, NULL);
-            g_object_unref (loader);
-            return NULL;
-        }
-
-        if (!gdk_pixbuf_loader_close (loader, error)) {
-            g_object_unref (loader);
-            return NULL;
-        }
-
-        pixbuf = gdk_pixbuf_loader_get_pixbuf (loader);
-
-        if (!pixbuf) {
-            g_object_unref (loader);
-            g_set_error (error,
-                         GDK_PIXBUF_ERROR,
-                         GDK_PIXBUF_ERROR_FAILED,
-                         _
-                         ("Failed to load image '%s': reason not known, probably a corrupt image file"),
-                         href);
-            return NULL;
-        }
-
-        g_object_ref (pixbuf);
+    res = gdk_pixbuf_loader_write (loader, arr->data, arr->len, error);
+    g_byte_array_free (arr, TRUE);
 
+    if (!res) {
+        gdk_pixbuf_loader_close (loader, NULL);
         g_object_unref (loader);
-
-        return pixbuf;
+        return NULL;
     }
 
-    return NULL;
+    if (!gdk_pixbuf_loader_close (loader, error)) {
+        g_object_unref (loader);
+        return NULL;
+    }
+
+    pixbuf = gdk_pixbuf_loader_get_pixbuf (loader);
+
+    if (!pixbuf) {
+        g_object_unref (loader);
+        g_set_error (error,
+                     GDK_PIXBUF_ERROR,
+                     GDK_PIXBUF_ERROR_FAILED,
+                      _("Failed to load image '%s': reason not known, probably a corrupt image file"),
+                      href);
+        return NULL;
+    }
+
+    surface = rsvg_cairo_surface_from_pixbuf (pixbuf);
+
+    g_object_unref (loader);
+
+    return surface;
 }
 
 void
@@ -275,8 +275,8 @@ rsvg_node_image_free (RsvgNode * self)
     rsvg_state_finalize (z->super.state);
     g_free (z->super.state);
     z->super.state = NULL;
-    if (z->img)
-        g_object_unref (z->img);
+    if (z->surface)
+        cairo_surface_destroy (z->surface);
     _rsvg_node_free(self);
 }
 
@@ -285,10 +285,10 @@ rsvg_node_image_draw (RsvgNode * self, RsvgDrawingCtx * ctx, int dominate)
 {
     RsvgNodeImage *z = (RsvgNodeImage *) self;
     unsigned int aspect_ratio = z->preserve_aspect_ratio;
-    GdkPixbuf *img = z->img;
     gdouble x, y, w, h;
+    cairo_surface_t *surface = z->surface;
 
-    if (img == NULL)
+    if (surface == NULL)
         return;
 
     x = _rsvg_css_normalize_length (&z->x, ctx, 'h');
@@ -304,10 +304,12 @@ rsvg_node_image_draw (RsvgNode * self, RsvgDrawingCtx * ctx, int dominate)
         rsvg_add_clipping_rect (ctx, x, y, w, h);
     }
 
-    rsvg_preserve_aspect_ratio (aspect_ratio, (double) gdk_pixbuf_get_width (img),
-                                (double) gdk_pixbuf_get_height (img), &w, &h, &x, &y);
+    rsvg_preserve_aspect_ratio (aspect_ratio, 
+                                (double) cairo_image_surface_get_width (surface),
+                                (double) cairo_image_surface_get_height (surface), 
+                                &w, &h, &x, &y);
 
-    rsvg_render_image (ctx, img, x, y, w, h);
+    rsvg_render_surface (ctx, surface, x, y, w, h);
 
     rsvg_pop_discrete_layer (ctx);
 }
@@ -330,11 +332,13 @@ rsvg_node_image_set_atts (RsvgNode * self, RsvgHandle * ctx, RsvgPropertyBag * a
         /* path is used by some older adobe illustrator versions */
         if ((value = rsvg_property_bag_lookup (atts, "path"))
             || (value = rsvg_property_bag_lookup (atts, "xlink:href"))) {
-            image->img = rsvg_pixbuf_new_from_href (value, rsvg_handle_get_base_uri (ctx), NULL);
+            image->surface = rsvg_cairo_surface_new_from_href (value, 
+                                                               rsvg_handle_get_base_uri (ctx), 
+                                                               NULL);
 
-            if (!image->img) {
+            if (!image->surface) {
 #ifdef G_ENABLE_DEBUG
-                g_warning (_("Couldn't load image: %s\n"), value);
+                g_warning ("Couldn't load image: %s\n", value);
 #endif
             }
         }
@@ -358,7 +362,7 @@ rsvg_new_image (void)
     image = g_new (RsvgNodeImage, 1);
     _rsvg_node_init (&image->super, RSVG_NODE_TYPE_IMAGE);
     g_assert (image->super.state);
-    image->img = NULL;
+    image->surface = NULL;
     image->preserve_aspect_ratio = RSVG_ASPECT_RATIO_XMID_YMID;
     image->x = image->y = image->w = image->h = _rsvg_css_parse_length ("0");
     image->super.free = rsvg_node_image_free;
